# -*- coding: utf-8 -*-

import smtplib
import logging
import traceback

from django.utils import timezone as tz
from django.conf import settings
from django.contrib.sites.models import Site

from crma.models import EmailScheduler, ST_SENT, ST_ERROR
from crma.mails import MailBuilder, BadEmailAddress
from crma.signals import email_sent, emails_sent
from django.core import mail


logger = logging.getLogger(__name__)


class PrepareError(Exception):
    pass


class EmailStream(object):
    @classmethod
    def from_queryset(cls, qs):
        pks = list(qs.values_list('pk', flat=True))
        return cls(qs.model.objects.filter(pk__in=pks).order_by('pk'))

    def __init__(self, qs):
        self.qs = qs

    def __len__(self):
        return self.qs.count()

    def __bool__(self):
        return len(self) > 0

    __nonzero__ = __bool__

    def __iter__(self):
        return self._iterate(self.qs)

    def __getitem__(self, slice_):
        return self._iterate(self.qs.__getitem__(slice_))

    def _iterate(self, qs):
        for email_scheduler in qs.iterator():
            sm = SendableMessage(email_scheduler)
            try:
                sm.prepare()
            except Exception:
                raise PrepareError(sm)
            yield sm


class ConnectionManager(object):
    def __init__(self, connection, messages, max_by_connection=0):
        self.connection = connection
        self.messages = messages
        self.max_by_connection = max_by_connection

    def __call__(self):
        index = 0
        remaining = len(self.messages)

        while index < remaining:
            if self.max_by_connection > 0:
                messages = self.messages[index:index + self.max_by_connection]
            else:
                messages = self.messages[index:]

            messages = iter(messages)
            prev_index = index
            try:
                self.connection.open()
                while True:
                    try:
                        # A PrepareError can happen in the next call
                        message = next(messages)
                        message.send(self.connection)
                    except StopIteration:
                        # Exhausted the messages, break
                        break
                    except smtplib.SMTPServerDisconnected:
                        # SMTP Disconnect, raise, the next try will catch it and retry
                        raise
                    except PrepareError as pe:
                        # Message failed to prepare, it has been marked in error, go forward
                        logger.exception('Cannot prepare %s', pe)
                        index += 1
                    except Exception as e:
                        # Message failed to be sent, it has been marked in error, go forward
                        logger.warning('Got an error after %s mails', index)
                        logger.warning('Cannot send %s: %s(%s)', message, type(e).__name__, e)
                        index += 1
                    else:
                        # go forward,it was sent
                        index += 1
            except smtplib.SMTPServerDisconnected:
                logger.warning('Disconnected after %s mails', index)
                continue
            finally:
                self.connection.close()
                if index == prev_index:
                    # Did not sent one mail, break
                    break


class SendableMessage(object):
    def __init__(self, email_scheduler):
        self.email_scheduler = email_scheduler
        self.message = None

    def __repr__(self):
        return 'M<{!r}>'.format(self.email_scheduler)

    def __eq__(self, other):
        if not isinstance(other, SendableMessage):
            return NotImplemented
        return self.email_scheduler == other.email_scheduler

    def _set_error(self, trace):
        self.email_scheduler.status = ST_ERROR
        self.email_scheduler.trace_error = traceback.format_exc()
        self.email_scheduler.save()

    def prepare(self):
        mb = MailBuilder(self.email_scheduler, Site.objects.get_current())
        try:
            message = mb.build_email()
        except BadEmailAddress as e:
            logging.debug('Skip mail to %s', e)
            self._set_error(e)
            raise
        except Exception as e:
            logger.exception('Error preparing %s', self)
            self._set_error(e)
            raise
        self.message = message
        return message

    def send(self, connection):
        assert self.message, 'Message has not been prepared'
        try:
            connection.send_messages([self.message])
        except smtplib.SMTPServerDisconnected:
            raise
        except smtplib.SMTPRecipientsRefused as e:
            logger.error('Error sending %s', self)
            self._set_error(e)
            raise
        except smtplib.SMTPResponseException as e:
            logger.error('Error sending %s', self)
            if e.smtp_code >= 500:
                self._set_error(e)
            raise
        except Exception as e:
            logger.exception('Error sending %s', self)
            self._set_error(e)
            raise
        else:
            self._set_sent()
        email_sent.send(sender=EmailScheduler, data={
            'id': self.email_scheduler.id,
            'email_scheduler': self.email_scheduler,
            'message': self.message,
        })

    def _set_sent(self):
        self.email_scheduler.status = ST_SENT
        self.email_scheduler.sent_time = tz.now()
        self.email_scheduler.save()


class Mailer(object):
    def __init__(self, date_fn=tz.now):
        self.date_fn = date_fn

    def __call__(self):
        # Emails to send
        now = self.date_fn()
        email_schedulers = EmailStream.from_queryset(EmailScheduler.objects.sendable(now))
        if not email_schedulers:
            logger.debug('Nothing to do')
            return

        logger.info('%s - %s prepared emails', now, len(email_schedulers))
        connection = mail.get_connection(getattr(settings, 'CRMA_EMAIL_BACKEND', None))
        max_mail_by_connection = getattr(settings, 'CRMA_MAX_MAILS_BY_CONNECTION', 0)
        sender = ConnectionManager(connection, email_schedulers, max_mail_by_connection)

        sender()
        emails_sent.send(sender=EmailScheduler, prepared_emails=email_schedulers)
